#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_lagging_classes.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
import re
import random
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event

# import trackme libs utils
from trackme_libs_utils import update_wildcard

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkLaggingClasses_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkLaggingClasses_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_lagging_classes(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_lagging_classes",
            "resource_group_desc": "Endpoints related to the management of lagging classes for splk-feeds components (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add new policy
    def post_lagging_classes_add(self, request_info, **kwargs):

        # Declare
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": "Error: tenant_id is required",
                        "status": 400,
                    }

                # component, valid values: dsm/dhm
                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm"):
                        return {
                            "payload": f'Invalid component="{component}", valid values: dsm/dhm',
                            "status": 500,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: component is required, valid values: dsm/dhm",
                        "status": 400,
                    }

                try:
                    name = resp_dict["name"]
                    if not len(name):
                        name = None
                except Exception as e:
                    name = None
                # if name is not set, generate a random name based on "rule-<random 6 digits id>"
                if name is None:
                    name = "rule-" + str(random.randint(100000, 999999))

                try:
                    level = resp_dict["level"]
                except Exception as e:
                    return {
                        "payload": "Error: level is required",
                        "status": 400,
                    }

                try:
                    value_lag = int(resp_dict["value_lag"])
                except Exception as e:
                    return {
                        "payload": "Error: value_lag is required and must be an integer",
                        "status": 400,
                    }

                try:
                    value_delay = int(resp_dict["value_delay"])
                except Exception as e:
                    return {
                        "payload": "Error: value_delay is required and must be an integer",
                        "status": 400,
                    }

                try:
                    value_future = int(resp_dict["value_future"])
                except Exception as e:
                    return {
                        "payload": "Error: value_future is required and must be an integer",
                        "status": 400,
                    }

                try:
                    allow_adaptive = resp_dict["allow_adaptive"]
                    # convert a boolean to a string
                    if allow_adaptive == 0 or allow_adaptive == "0":
                        allow_adaptive = "false"
                    elif allow_adaptive == 1 or allow_adaptive == "1":
                        allow_adaptive = "true"
                    if allow_adaptive not in ("true", "false"):
                        return {
                            "payload": "Error: allow_adaptive must be either true or false",
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: allow_adaptive is required",
                        "status": 400,
                    }

                try:
                    lag_alert_kpis = resp_dict["lag_alert_kpis"]
                    if lag_alert_kpis not in (
                        "all_kpis",
                        "lag_event_kpi",
                        "lag_ingestion_kpi",
                    ):
                        return {
                            "payload": "Error: lag_alert_kpis must be either all_kpis, lag_event_kpi or lag_ingestion_kpi",
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: lag_alert_kpis is required",
                        "status": 400,
                    }

                # optional: a comment for this entry
                try:
                    comment_value = resp_dict["comment"]
                    if len(comment_value) == 0:
                        comment_value = None
                except Exception as e:
                    comment_value = None

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new lagging class policy, it requires a POST call with the following data:",
                "resource_desc": "Create a new lagging class policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_add\" body=\"{'tenant_id':'mytenant','lagging_class_type':'events','name':'linux_secure','level':'sourcetype','object':'all','value_lag':'900','value_delay':'3600'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant idenfitier",
                        "component": "The component, valid values: dsm/dhm",
                        "name": "OPTIONAL: name of the lagging class, if not provided, a random name will be generated (rule-<random 6 digits id>)",
                        "level": "which level the lagging class is based on, valid otions are: object, alias, index, priority, sourcetype",
                        "value_lag": "the lagging value in seconds, an integer is expected",
                        "value_delay": "the delay value in seconds, an integer is expected",
                        "value_future": "the future tolerance value in seconds, an integer is expected",
                        "allow_adaptive": "allow adaptive lagging, valid options are: false / true / 0 / 1",
                        "lag_alert_kpis": "the lagging alerts KPIs, valid options are: all_kpis / lag_event_kpi / lag_ingestion_kpi",
                        "comment": "OPTIONAL: a comment for this blocklist, the comment will be stored and displayed when accessing records",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "$and": [
                {
                    "name": name,
                    "level": level,
                }
            ]
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = f"kv_trackme_{component}_lagging_classes_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            record = collection.data.query(query=json.dumps(query_string))[0]
            key = record.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            # conflict, the same entity exists already
            error_msg = f'tenant_id="{tenant_id}", conflict the same rule exists already, record="{json.dumps(record, indent=2)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

        else:

            # Perform and audit
            try:

                # define a value for rule_name, as: rule-<random 6 digits id>
                rule_name = "rule-" + str(random.randint(100000, 999999))

                # if the level contains wildcards, replace these with .* to interpret this as a regex automatically
                name = update_wildcard(name)

                record_class = {
                    "rule_name": rule_name,
                    "name": name,
                    "level": level,
                    "value_lag": str(value_lag),
                    "value_delay": str(value_delay),
                    "value_future": str(value_future),
                    "allow_adaptive": allow_adaptive,
                    "lag_alert_kpis": lag_alert_kpis,
                    "mtime": time.time(),
                }

                # Define if the pattern is a regex and store this information in the collection
                r = re.match("[\\\\|\\?|\\$|\\^|\\[|\\]|\\{|\\}|\\+]", name)
                r2 = re.findall("\\.[\\*|\\+]", name)

                if r or r2:
                    is_rex = "true"
                else:
                    is_rex = "false"

                record_class["is_rex"] = is_rex

                if comment_value:
                    record_class["comment"] = comment_value

                # Insert the record
                collection.data.insert(json.dumps(record_class))

                # Get record
                record = json.dumps(
                    collection.data.query(query=json.dumps(query_string)), indent=1
                )

                # Audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "create lagging class",
                        str(name),
                        "lagging_class",
                        str(record),
                        "The lagging class was created successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                logging.info("success for record=" + str(record))
                return {"payload": str(record), "status": 200}

            except Exception as e:
                response = {
                    "action": "failure",
                    "response": f'an exception was encountered, exception="{str(e)}"',
                }
                logging.error(json.dumps(response))
                return {"payload": response, "status": 500}

    # Add new policy
    def post_lagging_classes_metrics_add(self, request_info, **kwargs):

        # Declare
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": "Error: tenant_id is required",
                        "status": 400,
                    }

                try:
                    metric_category = resp_dict["metric_category"]
                except Exception as e:
                    return {
                        "payload": "Error: metric_category is required",
                        "status": 400,
                    }

                try:
                    metric_max_lag_allowed = int(resp_dict["metric_max_lag_allowed"])
                except Exception as e:
                    return {
                        "payload": "Error: metric_max_lag_allowed is required and must be an integer",
                        "status": 400,
                    }

                # optional: a comment for this entry
                try:
                    comment_value = resp_dict["comment"]
                    if len(comment_value) == 0:
                        comment_value = None
                except Exception as e:
                    comment_value = None

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new tag policy, it requires a POST call with the following data:",
                "resource_desc": "Create a new lagging class policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_metrics_add\" body=\"{'tenant_id':'mytenant','metric_category':'spl','metric_max_lag_allowed':'900'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "metric_category": "name of the metric category",
                        "metric_max_lag_allowed": "the lagging value in seconds, an integer is expected",
                        "comment": "OPTIONAL: a comment for this blocklist, the comment will be stored and displayed when accessing records",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = "kv_trackme_mhm_lagging_classes_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            # Define the KV query
            query_string = {
                "metric_category": metric_category,
            }

            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # conflict, the same entity exists already
                logging.error(
                    f'tenant_id="{tenant_id}", conflict the same object exists already, metric_category="{metric_category}"'
                )
                return {
                    "payload": f'tenant_id="{tenant_id}", conflict the same object exists already, metric_category="{metric_category}"',
                    "status": 500,
                }

            else:
                # This record does not exist yet

                # define a value for rule_name, as: rule-<random 6 digits id>
                rule_name = "rule-" + str(random.randint(100000, 999999))

                new_kvrecord = {
                    "rule_name": rule_name,
                    "metric_category": metric_category,
                    "metric_max_lag_allowed": metric_max_lag_allowed,
                    "mtime": time.time(),
                }

                if comment_value:
                    new_kvrecord["comment"] = comment_value

                # Perform and audit
                try:
                    # Insert the record
                    collection.data.insert(json.dumps(new_kvrecord))

                    # Audit
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "create metric sla policy",
                            str(metric_category),
                            "metric_sla_policy",
                            new_kvrecord,
                            "The metric SLA policy was created successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                except Exception as e:
                    response = {
                        "action": "failure",
                        "response": f'an exception was encountered, exception="{str(e)}"',
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

                response = {
                    "action": "success",
                    "response": "lagging class added successfully",
                    "record": new_kvrecord,
                }
                return {"payload": response, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": f'an exception was encountered, exception="{str(e)}"',
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Delete records from the collection
    def post_lagging_classes_del(self, request_info, **kwargs):

        # Declare
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": "Error: tenant_id is required",
                        "status": 400,
                    }

                # component, valid values: dsm/dhm
                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm"):
                        return {
                            "payload": f'Invalid component="{component}", valid values: dsm/dhm',
                            "status": 500,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: component is required, valid values: dsm/dhm",
                        "status": 400,
                    }

                try:
                    keys_list = resp_dict["keys_list"]
                    # Handle as a CSV list of keys
                    keys_list = keys_list.split(",")
                except Exception as e:
                    return {
                        "payload": "Error: keys_list is required",
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint deletes lagging classes, it requires a POST call with the following information:",
                "resource_desc": "Delete one or more lagging classes",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_del\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "component": "The component, valid options: dsm/dhm/mhm",
                        "keys_list": "List of record keys seperated by a comma of the records to be deleted from the collection",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = f"kv_trackme_{component}_lagging_classes_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # loop
        for item in keys_list:
            # Define the KV query
            query_string = {
                "_key": item,
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already

                # Store the record for audit purposes
                record = str(json.dumps(collection.data.query_by_id(key), indent=1))

                # Remove and audit
                try:
                    # Remove the record
                    collection.data.delete(json.dumps({"_key": key}))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "delete lagging class",
                            str(item),
                            "all",
                            str(record),
                            "The lagging class was deleted successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "delete lagging class",
                            str(item),
                            "all",
                            str(record),
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete lagging class",
                        str(item),
                        "all",
                        str(record),
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update records
    def post_lagging_classes_update(self, request_info, **kwargs):

        # Declare
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": "Error: tenant_id is required",
                        "status": 400,
                    }

                # component, valid values: dsm/dhm
                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm"):
                        return {
                            "payload": f'Invalid component="{component}", valid values: dsm/dhm',
                            "status": 500,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: component is required, valid values: dsm/dhm",
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm"):
                        return {
                            "payload": f'Invalid component="{component}"',
                            "status": 500,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: component is required",
                        "status": 400,
                    }

                try:
                    records_list = resp_dict["records_list"]
                    records_list = json.loads(records_list)
                except Exception as e:
                    return {
                        "payload": "Error: records_list is required",
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint updates records, it requires a POST call with the following information:",
                "resource_desc": "Update one or more lagging classes",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/post_lagging_classes_update\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "component": "The component, valid options: dsm/dhm/mhm",
                        "records_list": "JSON records to be updated",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = f"kv_trackme_{component}_lagging_classes_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # loop
        for item in records_list:

            # Define the KV query
            query_string = {
                "_key": item.get("_key"),
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already

                # Store the record for audit purposes
                record = str(json.dumps(collection.data.query_by_id(key), indent=1))

                # Update and audit
                try:
                    # Update the record

                    if component in ("dsm", "dhm"):

                        # get name
                        name = item.get("name")

                        # Define if the pattern is a regex and store this information in the collection
                        r = re.match("[\\\\|\\?|\\$|\\^|\\[|\\]|\\{|\\}|\\+]", name)
                        r2 = re.findall("\\.[\\*|\\+]", name)

                        if r or r2:
                            is_rex = "true"
                        else:
                            is_rex = "false"

                        # set record_class
                        record_class = {
                            "rule_name": item.get("rule_name"),
                            "name": name,
                            "level": item.get("level"),
                            "value_lag": item.get("value_lag"),
                            "value_delay": item.get("value_delay"),
                            "value_future": item.get("value_future"),
                            "allow_adaptive": item.get("allow_adaptive"),
                            "lag_alert_kpis": item.get("lag_alert_kpis"),
                            "is_rex": is_rex,
                            "comment": item.get("comment", ""),
                            "mtime": time.time(),
                        }

                        collection.data.update(
                            str(key),
                            json.dumps(record_class),
                        )

                    elif component in ("mhm"):
                        collection.data.update(
                            str(key),
                            json.dumps(
                                {
                                    "rule_name": item.get("rule_name"),
                                    "metric_category": item.get("metric_category"),
                                    "metric_max_lag_allowed": item.get(
                                        "metric_max_lag_allowed"
                                    ),
                                    "comment": item.get("comment", ""),
                                    "mtime": time.time(),
                                }
                            ),
                        )

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "update lagging classes",
                            str(item),
                            str(component),
                            json.dumps(item, indent=1),
                            "The lagging classs was updated successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "update lagging classes",
                            str(item),
                            str(component),
                            str(record),
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "update lagging classes",
                        str(item),
                        str(component),
                        str(record),
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}
