#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_lagging_classes.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel

# import trackme decision maker
from trackme_libs_decisionmaker import convert_epoch_to_datetime

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkLaggingClasses_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkLaggingClasses_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_lagging_classes(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_lagging_classes",
            "resource_group_desc": "Endpoints related to the management of lagging classes for splk-feeds components (read only operations)",
        }

        return {"payload": response, "status": 200}

    # get all records
    def post_lagging_classes_show(self, request_info, **kwargs):

        # Declare
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": "Error: tenant_id is required",
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm"):
                        return {
                            "payload": f'tenant_id="{tenant_id}", invalid option component="{component}", valid values are: dsm | dhm | mhm',
                            "status": 500,
                        }
                except Exception as e:
                    return {
                        "payload": "Error: component is required, valid values are: dsm | dhm | mhm",
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint retrieves all lagging classes, it requires a POST call with the following information:",
                "resource_desc": "Get lagging classes",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/lagging_classes_show\" body=\"{'tenant_id': 'mytenant', 'lagging_class_type': 'events'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "component": "The component, valid options: dsm/dhm/mhm",
                        "lagging_class_type": "The type of lagging classes, valid options are: events | metrics",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = (
                f"kv_trackme_{component}_lagging_classes_tenant_{tenant_id}"
            )
            collection = service.kvstore[collection_name]

            records = collection.data.query()
            results_records = []
            for item in records:
                mtime = item.get("mtime")
                if mtime:
                    mtime = convert_epoch_to_datetime(mtime)
                else:
                    mtime = "N/A"

                # records differ in structure
                if component in ("dsm", "dhm"):
                    results_records.append(
                        {
                            "_key": item.get("_key"),
                            "rule_name": item.get("rule_name"),
                            "level": item.get("level"),
                            "name": item.get("name"),
                            "value_lag": item.get("value_lag"),
                            "value_delay": item.get("value_delay"),
                            "value_future": item.get("value_future"),
                            "allow_adaptive": item.get("allow_adaptive"),
                            "lag_alert_kpis": item.get("lag_alert_kpis"),
                            "is_rex": item.get("is_rex", "false"),
                            "comment": item.get("comment", ""),
                            "mtime": mtime,
                        }
                    )

                elif component == "mhm":
                    results_records.append(
                        {
                            "_key": item.get("_key"),
                            "rule_name": item.get("rule_name"),
                            "metric_category": item.get("metric_category"),
                            "metric_max_lag_allowed": item.get(
                                "metric_max_lag_allowed"
                            ),
                            "is_rex": item.get("is_rex", "false"),
                            "comment": item.get("comment", ""),
                            "mtime": mtime,
                        }
                    )

            return {"payload": results_records, "status": 200}

        except Exception as e:
            error_msg = f"Error: {str(e)}"
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}
